@@ -1,7 +1,7 @@
 /*
  * Peony-Qt's Library
  *
- * Copyright (C) 2019, Tianjin KYLIN Information Technology Co., Ltd.
+ * Copyright (C) 2020, KylinSoft Co., Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -28,6 +28,7 @@
 #include "directory-view-container.h"
 
 #include "menu-plugin-manager.h"
+#include "file-info-job.h"
 #include "file-info.h"
 
 #include "directory-view-factory-manager.h"
@@ -39,6 +40,7 @@
 #include "file-operation-manager.h" //FileOpInfo
 
 #include "file-utils.h"
+#include "bookmark-manager.h"
 
 #include "volume-manager.h"
 
@@ -57,7 +59,6 @@
 #include <QFileInfo>
 #include <QDir>
 #include <QStandardPaths>
-#include <QFileIconProvider>
 
 #include <QLocale>
 #include <QStandardPaths>
@@ -100,6 +101,9 @@ void DirectoryViewMenu::fillActions()
     if (m_directory.startsWith("search://"))
         m_is_search = true;
 
+    if (m_directory.startsWith("burn://"))
+        m_is_cd = true;
+
     //add open actions
     auto openActions = constructOpenOpActions();
     if (!openActions.isEmpty())
@@ -182,6 +186,16 @@ const QList<QAction *> DirectoryViewMenu::constructOpenOpActions()
                 displayName = fontMetrics().elidedText(displayName, Qt::ElideRight, ELIDE_TEXT_LENGTH * charWidth);
             }
             if (info->isDir()) {
+                //add to bookmark option
+                l<<addAction(QIcon::fromTheme("bookmark-add-symbolic"), tr("Add to bookmark"));
+                connect(l.last(), &QAction::triggered, [=]() {
+                    //qDebug() <<"add to bookmark:" <<info->uri();
+                    auto bookmark = BookMarkManager::getInstance();
+                    if (bookmark->isLoaded()) {
+                        bookmark->addBookMark(info->uri());
+                    }
+                });
+
                 l<<addAction(QIcon::fromTheme("document-open-symbolic"), tr("&Open \"%1\"").arg(displayName));
                 connect(l.last(), &QAction::triggered, [=]() {
                     if (!m_top_window)
@@ -240,20 +254,42 @@ const QList<QAction *> DirectoryViewMenu::constructOpenOpActions()
             connect(l.last(), &QAction::triggered, [=]() {
                 qDebug()<<"triggered";
                 QStringList dirs;
-                QStringList files;
+                QMap<QString, QStringList> fileMap;
+                /**step 1: Categorize files according to type.
+                 * step 2: Open files in batches to avoid loss of asynchronous messages due to program startup.
+                **/
                 for (auto uri : m_selections) {
                     auto info = FileInfo::fromUri(uri);
                     if (info->isDir() || info->isVolume()) {
                         dirs<<uri;
                     } else {
-                        files<<uri;
+                        QString mimeType = info->mimeType();
+                        if (mimeType.isEmpty()) {
+                            FileInfoJob job(info);
+                            job.querySync();
+                            mimeType = info->mimeType();
+                        }
+
+                        QStringList list;
+                        if (fileMap.contains(mimeType)) {
+                            list = fileMap[mimeType];
+                            list << uri;
+                            fileMap.insert(mimeType, list);
+                        } else {
+                            list << uri;
+                            fileMap.insert(mimeType, list);
+                        }
                     }
                 }
                 if (!dirs.isEmpty())
                     m_top_window->addNewTabs(dirs);
-                if (!files.isEmpty()) {
-                    for (auto uri : files) {
-                        FileLaunchManager::openAsync(uri);
+
+                if(!fileMap.empty()) {
+                    QMap<QString, QStringList>::iterator iter = fileMap.begin();
+                    while (iter != fileMap.end())
+                    {
+                        FileLaunchManager::openAsync(iter.value());
+                        iter++;
                     }
                 }
             });
@@ -268,6 +304,13 @@ const QList<QAction *> DirectoryViewMenu::constructCreateTemplateActions()
     QList<QAction *> l;
     if (m_selections.isEmpty()) {
         auto createAction = new QAction(tr("&New..."), this);
+        if (m_is_cd) {
+            createAction->setEnabled(false);
+        }
+        if(m_directory.compare(QString::fromLocal8Bit("trash:///")) == 0)
+        {
+            createAction->setEnabled(false);
+        }
         l<<createAction;
         QMenu *subMenu = new QMenu(this);
         createAction->setMenu(subMenu);
@@ -277,19 +320,50 @@ const QList<QAction *> DirectoryViewMenu::constructCreateTemplateActions()
         QDir templateDir(g_get_user_special_dir(G_USER_DIRECTORY_TEMPLATES));
         auto templates = templateDir.entryList(QDir::AllEntries|QDir::NoDotAndDotDot);
         if (!templates.isEmpty()) {
-            QFileIconProvider p;
             for (auto t : templates) {
-                QFileInfo info(t);
-                QAction *action = new QAction(p.icon(info), info.baseName(), this);
+                QFileInfo qinfo(templateDir, t);
+                GFile *gtk_file = g_file_new_for_path(qinfo.filePath().toUtf8().data());
+                char *uri_str = g_file_get_uri(gtk_file);
+                std::shared_ptr<FileInfo> info = FileInfo::fromUri(uri_str);
+
+                QString mimeType = info->mimeType();
+                if (mimeType.isEmpty()) {
+                    FileInfoJob job(info);
+                    job.querySync();
+                    mimeType = info->mimeType();
+                }
+
+                QIcon tmpIcon;
+                GList *app_infos = g_app_info_get_recommended_for_type(mimeType.toUtf8().constData());
+                GList *l = app_infos;
+                QList<FileLaunchAction *> actions;
+                bool isOnlyUnref = false;
+                while (l) {
+                    auto app_info = static_cast<GAppInfo*>(l->data);
+                    if (!isOnlyUnref) {
+                        GThemedIcon *icon = G_THEMED_ICON(g_app_info_get_icon(app_info));
+                        const char * const * icon_names = g_themed_icon_get_names(icon);
+                        if (icon_names)
+                            tmpIcon = QIcon::fromTheme(*icon_names);
+                        if(!tmpIcon.isNull())
+                            isOnlyUnref = true;
+                    }
+                    g_object_unref(app_infos);
+                    l = l->next;
+                }
+
+                QAction *action = new QAction(tmpIcon, qinfo.baseName(), this);
                 connect(action, &QAction::triggered, [=]() {
                     CreateTemplateOperation op(m_directory, CreateTemplateOperation::Template, t);
-                    FileOperationErrorDialog dlg;
-                    connect(&op, &CreateTemplateOperation::errored, &dlg, &FileOperationErrorDialog::handleError);
+                    Peony::FileOperationErrorDialogConflict dlg;
+                    connect(&op, &Peony::FileOperation::errored, &dlg, &Peony::FileOperationErrorDialogConflict::handle);
                     op.run();
                     auto target = op.target();
                     m_uris_to_edit<<target;
                 });
                 subMenu->addAction(action);
+                g_free(uri_str);
+                g_object_unref(gtk_file);
             }
             subMenu->addSeparator();
         }
@@ -300,8 +374,8 @@ const QList<QAction *> DirectoryViewMenu::constructCreateTemplateActions()
         connect(actions.last(), &QAction::triggered, [=]() {
             //FileOperationUtils::create(m_directory);
             CreateTemplateOperation op(m_directory);
-            FileOperationErrorDialog dlg;
-            connect(&op, &CreateTemplateOperation::errored, &dlg, &FileOperationErrorDialog::handleError);
+            Peony::FileOperationErrorDialogConflict dlg;
+            connect(&op, &Peony::FileOperation::errored, &dlg, &Peony::FileOperationErrorDialogConflict::handle);
             op.run();
             auto targetUri = op.target();
             qDebug()<<"target:"<<targetUri;
@@ -312,8 +386,8 @@ const QList<QAction *> DirectoryViewMenu::constructCreateTemplateActions()
         connect(actions.last(), &QAction::triggered, [=]() {
             //FileOperationUtils::create(m_directory, nullptr, CreateTemplateOperation::EmptyFolder);
             CreateTemplateOperation op(m_directory, CreateTemplateOperation::EmptyFolder, tr("New Folder"));
-            FileOperationErrorDialog dlg;
-            connect(&op, &CreateTemplateOperation::errored, &dlg, &FileOperationErrorDialog::handleError);
+            Peony::FileOperationErrorDialogConflict dlg;
+            connect(&op, &Peony::FileOperation::errored, &dlg, &Peony::FileOperationErrorDialogConflict::handle);
             op.run();
             auto targetUri = op.target();
             qDebug()<<"target:"<<targetUri;
@@ -362,8 +436,8 @@ const QList<QAction *> DirectoryViewMenu::constructViewOpActions()
 
         QList<QAction *> tmp;
         tmp<<sortTypeMenu->addAction(tr("Name"));
-        tmp<<sortTypeMenu->addAction(tr("File Type"));
         tmp<<sortTypeMenu->addAction(tr("File Size"));
+        tmp<<sortTypeMenu->addAction(tr("File Type"));
         tmp<<sortTypeMenu->addAction(tr("Modified Date"));
         int sortType = m_view->getSortType();
         if (sortType >= 0) {
@@ -450,10 +524,15 @@ const QList<QAction *> DirectoryViewMenu::constructFileOpActions()
             connect(l.last(), &QAction::triggered, [=]() {
                 ClipboardUtils::setClipboardFiles(m_selections, true);
             });
-            l<<addAction(QIcon::fromTheme("edit-delete-symbolic"), tr("&Delete"));
+            l<<addAction(QIcon::fromTheme("edit-delete-symbolic"), tr("&Delete to trash"));
             connect(l.last(), &QAction::triggered, [=]() {
                 FileOperationUtils::trash(m_selections, true);
             });
+            //add delete forever option
+            l<<addAction(QIcon::fromTheme("edit-clear-symbolic"), tr("Delete forever"));
+            connect(l.last(), &QAction::triggered, [=]() {
+                FileOperationUtils::executeRemoveActionWithDialog(m_selections);
+            });
             if (m_selections.count() == 1) {
                 l<<addAction(QIcon::fromTheme("document-edit-symbolic"), tr("&Rename"));
                 connect(l.last(), &QAction::triggered, [=]() {
@@ -474,6 +553,24 @@ const QList<QAction *> DirectoryViewMenu::constructFileOpActions()
         }
     }
 
+    //select all and reverse select
+    if (m_selections.isEmpty())
+    {
+        l<<addAction(tr("Select &All"));
+        connect(l.last(), &QAction::triggered, [=]() {
+            //qDebug() << "select all";
+            m_view->invertSelections();
+        });
+    }
+    else
+    {
+        l<<addAction(tr("Reverse Select"));
+        connect(l.last(), &QAction::triggered, [=]() {
+            //qDebug() << "Reverse select";
+            m_view->invertSelections();
+        });
+    }
+
     return l;
 }
 
@@ -482,7 +579,7 @@ const QList<QAction *> DirectoryViewMenu::constructFilePropertiesActions()
     QList<QAction *> l;
 
     if (!m_is_search) {
-        l<<addAction(QIcon::fromTheme("preview-file"), tr("&Properties"));
+        l<<addAction(QIcon::fromTheme("preview-file"), tr("P&roperties"));
         connect(l.last(), &QAction::triggered, [=]() {
             //FIXME:
             if (m_selections.isEmpty()) {
@@ -498,7 +595,7 @@ const QList<QAction *> DirectoryViewMenu::constructFilePropertiesActions()
             }
         });
     } else if (m_selections.count() == 1) {
-        l<<addAction(QIcon::fromTheme("preview-file"), tr("&Properties"));
+        l<<addAction(QIcon::fromTheme("preview-file"), tr("P&roperties"));
         connect(l.last(), &QAction::triggered, [=]() {
             PropertiesWindow *p = new PropertiesWindow(m_selections);
             p->setAttribute(Qt::WA_DeleteOnClose);
@@ -546,8 +643,8 @@ const QList<QAction *> DirectoryViewMenu::constructTrashActions()
             l.last()->setEnabled(!isTrashEmpty);
             connect(l.last(), &QAction::triggered, [=]() {
                 auto result = QMessageBox::question(nullptr, tr("Delete Permanently"), tr("Are you sure that you want to delete these files? "
-                                                    "Once you start a deletion, the files deleting will never be "
-                                                    "restored again."));
+                                                                                          "Once you start a deletion, the files deleting will never be "
+                                                                                          "restored again."));
                 if (result == QMessageBox::Yes) {
                     auto uris = m_top_window->getCurrentAllFileUris();
                     FileOperationUtils::remove(uris);
@@ -565,8 +662,8 @@ const QList<QAction *> DirectoryViewMenu::constructTrashActions()
             l<<addAction(QIcon::fromTheme("window-close-symbolic"), tr("&Delete"));
             connect(l.last(), &QAction::triggered, [=]() {
                 auto result = QMessageBox::question(nullptr, tr("Delete Permanently"), tr("Are you sure that you want to delete these files? "
-                                                    "Once you start a deletion, the files deleting will never be "
-                                                    "restored again."));
+                                                                                          "Once you start a deletion, the files deleting will never be "
+                                                                                          "restored again."));
                 if (result == QMessageBox::Yes) {
                     FileOperationUtils::remove(m_selections);
                 }
